@@ -34,6 +34,7 @@
 
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <sys/stat.h>
 
 #include <assert.h>
 #include <errno.h>
@@ -53,6 +54,8 @@ struct _adcli_conn_ctx {
 	char *computer_password;
 	char *login_ccache_name;
 	int login_ccache_name_is_krb5;
+	char *login_keytab_name;
+	int login_keytab_name_is_krb5;
 	adcli_login_type login_type;
 	int logins_allowed;
 
@@ -79,6 +82,7 @@ struct _adcli_conn_ctx {
 	int ldap_authenticated;
 	krb5_context k5;
 	krb5_ccache ccache;
+	krb5_keytab keytab;
 };
 
 static adcli_result
@@ -180,11 +184,45 @@ ensure_domain_and_host (adcli_result res,
 	return ADCLI_SUCCESS;
 }
 
+char *
+_adcli_calc_netbios_name (const char *host_fqdn)
+{
+	const char *dom;
+	char *computer_name;
+
+	/* Use the FQDN minus the last part */
+	dom = strchr (host_fqdn, '.');
+
+	/* If dot is first then fail */
+	if (dom == host_fqdn) {
+		_adcli_err ("Couldn't determine the computer account name from host name: %s",
+		            host_fqdn);
+		return NULL;
+
+	} else if (dom == NULL) {
+		computer_name = strdup (host_fqdn);
+		return_val_if_fail (computer_name != NULL, NULL);
+
+	} else {
+		computer_name = strndup (host_fqdn, dom - host_fqdn);
+		return_val_if_fail (computer_name != NULL, NULL);
+	}
+
+	_adcli_str_up (computer_name);
+	if (strlen (computer_name) > 15) {
+		computer_name[15] = 0;
+		_adcli_info ("Truncated computer account name from fqdn: %s", computer_name);
+	} else {
+		_adcli_info ("Calculated computer account name from fqdn: %s", computer_name);
+	}
+
+	return computer_name;
+}
+
 static adcli_result
 ensure_computer_name (adcli_result res,
                       adcli_conn *conn)
 {
-	const char *dom;
 
 	if (res != ADCLI_SUCCESS)
 		return res;
@@ -196,27 +234,10 @@ ensure_computer_name (adcli_result res,
 
 	assert (conn->host_fqdn != NULL);
 
-	/* Use the FQDN minus the last part */
-	dom = strchr (conn->host_fqdn, '.');
-
-	/* If dot is first then fail */
-	if (dom == conn->host_fqdn) {
-		_adcli_err ("Couldn't determine the computer account name from host name: %s",
-		            conn->host_fqdn);
+	conn->computer_name = _adcli_calc_netbios_name (conn->host_fqdn);
+	if (conn->computer_name == NULL)
 		return ADCLI_ERR_CONFIG;
 
-	} else if (dom == NULL) {
-		conn->computer_name = strdup (conn->host_fqdn);
-		return_unexpected_if_fail (conn->computer_name != NULL);
-
-	} else {
-		conn->computer_name = strndup (conn->host_fqdn, dom - conn->host_fqdn);
-		return_unexpected_if_fail (conn->computer_name != NULL);
-	}
-
-	_adcli_str_up (conn->computer_name);
-
-	_adcli_info ("Calculated computer account name from fqdn: %s", conn->computer_name);
 	return ADCLI_SUCCESS;
 }
 
@@ -242,27 +263,6 @@ ensure_domain_realm (adcli_result res,
 	return ADCLI_SUCCESS;
 }
 
-static adcli_result
-ensure_k5_ctx (adcli_conn *conn)
-{
-	krb5_error_code code;
-
-	if (conn->k5)
-		return ADCLI_SUCCESS;
-
-	code = krb5_init_context (&conn->k5);
-	if (code == ENOMEM) {
-		return_unexpected_if_reached ();
-
-	} else if (code != 0) {
-		_adcli_err ("Failed to create kerberos context: %s",
-		            krb5_get_error_message (NULL, code));
-		return ADCLI_ERR_UNEXPECTED;
-	}
-
-	return ADCLI_SUCCESS;
-}
-
 static adcli_result
 ensure_user_password (adcli_conn *conn)
 {
@@ -355,9 +355,11 @@ setup_krb5_conf_snippet (adcli_conn *conn)
 {
 	char *filename;
 	char *snippet;
+	char *controller;
 	int errn;
 	int ret;
 	int fd;
+	mode_t old_mask;
 
 	if (!conn->krb5_conf_dir)
 		return ADCLI_SUCCESS;
@@ -371,6 +373,15 @@ setup_krb5_conf_snippet (adcli_conn *conn)
 	if (asprintf (&filename, "%s/adcli-krb5-conf-XXXXXX", conn->krb5_conf_dir) < 0)
 		return_unexpected_if_reached ();
 
+	if (strchr (conn->domain_controller, ':')) {
+		if (asprintf (&controller, "[%s]", conn->domain_controller) < 0)
+			controller = NULL;
+	} else {
+		controller = strdup (conn->domain_controller);
+	}
+
+	return_unexpected_if_fail (controller != NULL);
+
 	if (asprintf (&snippet, "[realms]\n"
 	                        "  %s = {\n"
 	                        "    kdc = %s:88\n"
@@ -380,12 +391,14 @@ setup_krb5_conf_snippet (adcli_conn *conn)
 	                        "[domain_realm]\n"
 	                        "  %s = %s\n"
 	                        "  %s = %s\n",
-	              conn->domain_realm, conn->domain_controller, conn->domain_controller, conn->domain_controller,
+	              conn->domain_realm, controller, controller, controller,
 	              conn->canonical_host, conn->domain_realm,
 	              conn->domain_controller, conn->domain_realm) < 0)
 		return_unexpected_if_reached ();
 
+	old_mask = umask (0177);
 	fd = mkstemp (filename);
+	umask (old_mask);
 	if (fd < 0) {
 		_adcli_warn ("Couldn't create krb5.conf snippet file in: %s: %s",
 		             conn->krb5_conf_dir, strerror (errno));
@@ -412,6 +425,7 @@ setup_krb5_conf_snippet (adcli_conn *conn)
 		}
 	}
 
+	free (controller);
 	free (snippet);
 
 	/* This shouldn't stop joining */
@@ -483,24 +497,30 @@ _adcli_kinit_computer_creds (adcli_conn *conn,
 	 * explicitly requested.
 	 */
 
-	if (!password && conn->password_func &&
-	    conn->logins_allowed == ADCLI_LOGIN_COMPUTER_ACCOUNT) {
-		new_password = (conn->password_func) (ADCLI_LOGIN_COMPUTER_ACCOUNT,
-		                                      sam, 0, conn->password_data);
-		password = new_password;
-	}
+	if (conn->keytab) {
+		code = krb5_get_init_creds_keytab (k5, creds, principal, conn->keytab,
+		                                   0, (char *)in_tkt_service, opt);
 
-	if (password == NULL) {
-		new_password = _adcli_calc_reset_password (conn->computer_name);
-		password = new_password;
-	}
+	} else {
+		if (!password && conn->password_func &&
+		    conn->logins_allowed == ADCLI_LOGIN_COMPUTER_ACCOUNT) {
+			new_password = (conn->password_func) (ADCLI_LOGIN_COMPUTER_ACCOUNT,
+			                                      sam, 0, conn->password_data);
+			password = new_password;
+		}
 
-	code = krb5_get_init_creds_password (k5, creds, principal, (char *)password,
-	                                     null_prompter, NULL, 0, (char *)in_tkt_service, opt);
+		if (password == NULL) {
+			new_password = _adcli_calc_reset_password (conn->computer_name);
+			password = new_password;
+		}
 
-	if (code == 0 && new_password) {
-		_adcli_password_free (conn->computer_password);
-		conn->computer_password = new_password;
+		code = krb5_get_init_creds_password (k5, creds, principal, (char *)password,
+		                                     null_prompter, NULL, 0, (char *)in_tkt_service, opt);
+
+		if (code == 0 && new_password) {
+			_adcli_password_free (conn->computer_password);
+			conn->computer_password = new_password;
+		}
 	}
 
 	krb5_free_principal (k5, principal);
@@ -622,10 +642,6 @@ prep_kerberos_and_kinit (adcli_conn *conn)
 	krb5_ccache ccache;
 	adcli_result res;
 
-	res = ensure_k5_ctx (conn);
-	if (res != ADCLI_SUCCESS)
-		return res;
-
 	if (conn->login_ccache_name != NULL) {
 		if (!conn->ccache) {
 
@@ -661,6 +677,25 @@ prep_kerberos_and_kinit (adcli_conn *conn)
 		return ADCLI_SUCCESS;
 	}
 
+	if (conn->login_keytab_name != NULL) {
+		if (!conn->keytab) {
+			res = _adcli_krb5_open_keytab (conn->k5, conn->login_keytab_name, &conn->keytab);
+			if (res != ADCLI_SUCCESS) {
+				if (res == ADCLI_ERR_FAIL)
+					res = ADCLI_ERR_CONFIG;
+				return res;
+			}
+
+			if (strcmp (conn->login_keytab_name, "") == 0) {
+				free (conn->login_keytab_name);
+				conn->login_keytab_name = malloc (MAX_KEYTAB_NAME_LEN);
+				code = krb5_kt_get_name (conn->k5, conn->keytab,
+				                         conn->login_keytab_name, MAX_KEYTAB_NAME_LEN);
+				conn->login_keytab_name_is_krb5 = 1;
+				return_unexpected_if_fail (code == 0);
+			}
+		}
+	}
 
 	/* Initialize the credential cache */
 	code = krb5_cc_new_unique (conn->k5, "MEMORY", NULL, &ccache);
@@ -692,6 +727,9 @@ prep_kerberos_and_kinit (adcli_conn *conn)
 		conn->ccache = ccache;
 		conn->login_ccache_name_is_krb5 = 1;
 		ccache = NULL;
+		res = ADCLI_SUCCESS;
+	} else {
+		res = ADCLI_ERR_FAIL;
 	}
 
 	if (ccache != NULL)
@@ -733,6 +771,7 @@ connect_to_address (const char *host,
 	}
 
 	for (ai = res; ai != NULL; ai = ai->ai_next) {
+		/* coverity[overwrite_var] */
 		sock = socket (ai->ai_family, ai->ai_socktype, ai->ai_protocol);
 		if (sock < 0) {
 			error = errno;
@@ -758,6 +797,7 @@ connect_to_address (const char *host,
 		_adcli_err ("Couldn't connect to host: %s: %s", host, strerror (error));
 
 	freeaddrinfo (res);
+	/* coverity[leaked_handle] - the socket is carried inside the ldap struct */
 	return ldap;
 }
 
@@ -951,7 +991,7 @@ authenticate_to_directory (adcli_conn *conn)
 {
 	OM_uint32 status;
 	OM_uint32 minor;
-	int opt;
+	ber_len_t ssf;
 	int ret;
 
 	if (conn->ldap_authenticated)
@@ -965,8 +1005,8 @@ authenticate_to_directory (adcli_conn *conn)
 	return_unexpected_if_fail (status == 0);
 
 	/* Clumsily tell ldap + cyrus-sasl that we want encryption */
-	opt = 1;
-	ret = ldap_set_option (conn->ldap, LDAP_OPT_X_SASL_SSF_MIN, &opt);
+	ssf = 1;
+	ret = ldap_set_option (conn->ldap, LDAP_OPT_X_SASL_SSF_MIN, &ssf);
 	return_unexpected_if_fail (ret == 0);
 
 	ret = ldap_sasl_interactive_bind_s (conn->ldap, NULL, "GSSAPI", NULL, NULL,
@@ -1044,6 +1084,10 @@ conn_clear_state (adcli_conn *conn)
 		krb5_cc_close (conn->k5, conn->ccache);
 	conn->ccache = NULL;
 
+	if (conn->keytab)
+		krb5_kt_close (conn->k5, conn->keytab);
+	conn->keytab = NULL;
+
 	if (conn->k5)
 		krb5_free_context (conn->k5);
 	conn->k5 = NULL;
@@ -1088,6 +1132,11 @@ adcli_conn_connect (adcli_conn *conn)
 	if (res != ADCLI_SUCCESS)
 		return res;
 
+	return_unexpected_if_fail (conn->k5 == NULL);
+	res = _adcli_krb5_init_context (&conn->k5);
+	if (res != ADCLI_SUCCESS)
+		return res;
+
 	/* Login with admin credentials now, setup login ccache */
 	res = prep_kerberos_and_kinit (conn);
 	if (res != ADCLI_SUCCESS)
@@ -1289,10 +1338,7 @@ krb5_context
 adcli_conn_get_krb5_context (adcli_conn *conn)
 {
 	return_val_if_fail (conn != NULL, NULL);
-
-	if (ensure_k5_ctx (conn) != ADCLI_SUCCESS)
-		return NULL;
-
+	return_val_if_fail (conn->k5 != NULL, NULL);
 	return conn->k5;
 }
 
@@ -1406,6 +1452,42 @@ adcli_conn_set_login_ccache_name (adcli_conn *conn,
 	conn->login_ccache_name_is_krb5 = 0;
 }
 
+const char *
+adcli_conn_get_login_keytab_name (adcli_conn *conn)
+{
+	return_val_if_fail (conn != NULL, NULL);
+	return conn->login_keytab_name;
+}
+
+void
+adcli_conn_set_login_keytab_name (adcli_conn *conn,
+                                  const char *ktname)
+{
+	char *newval = NULL;
+
+	return_if_fail (conn != NULL);
+
+	if (ktname) {
+		newval = strdup (ktname);
+		return_if_fail (newval != NULL);
+	}
+
+	if (conn->login_keytab_name) {
+		if (conn->login_keytab_name_is_krb5)
+			krb5_free_string (conn->k5, conn->login_keytab_name);
+		else
+			free (conn->login_keytab_name);
+	}
+
+	if (conn->keytab) {
+		krb5_kt_close (conn->k5, conn->keytab);
+		conn->keytab = NULL;
+	}
+
+	conn->login_keytab_name = newval;
+	conn->login_keytab_name_is_krb5 = 0;
+}
+
 const char *
 adcli_conn_get_default_naming_context (adcli_conn *conn)
 {
