@@ -129,6 +129,11 @@ const char *        adcli_conn_get_login_ccache_name (adcli_conn *conn);
 void                adcli_conn_set_login_ccache_name (adcli_conn *conn,
                                                       const char *ccname);
 
+const char *        adcli_conn_get_login_keytab_name (adcli_conn *conn);
+
+void                adcli_conn_set_login_keytab_name (adcli_conn *conn,
+                                                      const char *ktname);
+
 const char *        adcli_conn_get_default_naming_context (adcli_conn *conn);
 
 const char *        adcli_conn_get_krb5_conf_dir     (adcli_conn *conn);
