@@ -29,6 +29,7 @@
 #include <assert.h>
 #include <err.h>
 #include <stdio.h>
+#include <errno.h>
 
 static void
 dump_details (adcli_conn *conn,
@@ -103,6 +104,7 @@ typedef enum {
 	opt_os_version,
 	opt_os_service_pack,
 	opt_user_principal,
+	opt_computer_password_lifetime,
 } Option;
 
 static adcli_tool_desc common_usages[] = {
@@ -128,6 +130,7 @@ static adcli_tool_desc common_usages[] = {
 	{ opt_os_version, "the computer operating system version", },
 	{ opt_os_service_pack, "the computer operating system service pack", },
 	{ opt_user_principal, "add an authentication principal to the account", },
+	{ opt_computer_password_lifetime, "lifetime of the host accounts password in days", },
 	{ opt_no_password, "don't prompt for or read a password" },
 	{ opt_prompt_password, "prompt for a password if necessary" },
 	{ opt_stdin_password, "read a password from stdin (until EOF) if\n"
@@ -151,6 +154,8 @@ parse_option (Option opt,
 	static int no_password = 0;
 	static int prompt_password = 0;
 	static int stdin_password = 0;
+	char *endptr;
+	unsigned int lifetime;
 
 	switch (opt) {
 	case opt_login_ccache:
@@ -244,6 +249,18 @@ parse_option (Option opt,
 		else
 			adcli_enroll_auto_user_principal (enroll);
 		return;
+	case opt_computer_password_lifetime:
+		errno = 0;
+		lifetime = strtoul (optarg, &endptr, 10);
+		if (errno != 0 || *endptr != '\0' || endptr == optarg) {
+			errx (EUSAGE,
+			      "failure to parse value '%s' of option 'computer-password-lifetime'; "
+			      "expecting non-negative integer indicating the lifetime in days",
+			      optarg);
+		}
+
+		adcli_enroll_set_computer_password_lifetime (enroll, lifetime);
+		return;
 	case opt_verbose:
 		return;
 
@@ -339,6 +356,7 @@ adcli_tool_computer_join (adcli_conn *conn,
 		case ':':
 			adcli_tool_usage (options, usages);
 			adcli_tool_usage (options, common_usages);
+			adcli_enroll_unref (enroll);
 			return opt == 'h' ? 0 : 2;
 		default:
 			parse_option ((Option)opt, optarg, conn, enroll);
@@ -378,6 +396,110 @@ adcli_tool_computer_join (adcli_conn *conn,
 	return 0;
 }
 
+int
+adcli_tool_computer_update (adcli_conn *conn,
+		            int argc,
+                            char *argv[])
+{
+	adcli_enroll_flags flags = ADCLI_ENROLL_ALLOW_OVERWRITE;
+	adcli_enroll *enroll;
+	adcli_result res;
+	int show_password = 0;
+	int details = 0;
+	const char *ktname;
+	int opt;
+
+	struct option options[] = {
+		{ "domain", required_argument, NULL, opt_domain },
+		{ "domain-controller", required_argument, NULL, opt_domain_controller },
+		{ "host-fqdn", required_argument, 0, opt_host_fqdn },
+		{ "computer-name", required_argument, 0, opt_computer_name },
+		{ "host-keytab", required_argument, 0, opt_host_keytab },
+		{ "login-ccache", optional_argument, NULL, opt_login_ccache },
+		{ "service-name", required_argument, NULL, opt_service_name },
+		{ "os-name", required_argument, NULL, opt_os_name },
+		{ "os-version", required_argument, NULL, opt_os_version },
+		{ "os-service-pack", optional_argument, NULL, opt_os_service_pack },
+		{ "user-principal", optional_argument, NULL, opt_user_principal },
+		{ "computer-password-lifetime", optional_argument, NULL, opt_computer_password_lifetime },
+		{ "show-details", no_argument, NULL, opt_show_details },
+		{ "show-password", no_argument, NULL, opt_show_password },
+		{ "verbose", no_argument, NULL, opt_verbose },
+		{ "help", no_argument, NULL, 'h' },
+		{ 0 },
+	};
+
+	static adcli_tool_desc usages[] = {
+		{ 0, "usage: adcli update" },
+		{ 0 },
+	};
+
+	enroll = adcli_enroll_new (conn);
+	if (enroll == NULL)
+		errx (-1, "unexpected memory problems");
+
+	while ((opt = adcli_tool_getopt (argc, argv, options)) != -1) {
+		switch (opt) {
+		case opt_show_details:
+			details = 1;
+			break;
+		case opt_show_password:
+			show_password = 1;
+			break;
+		case 'h':
+		case '?':
+		case ':':
+			adcli_tool_usage (options, usages);
+			adcli_tool_usage (options, common_usages);
+			adcli_enroll_unref (enroll);
+			return opt == 'h' ? 0 : 2;
+		default:
+			parse_option ((Option)opt, optarg, conn, enroll);
+			break;
+		}
+	}
+
+	argc -= optind;
+	argv += optind;
+
+	if (adcli_conn_get_login_ccache_name (conn) == NULL) {
+		/* Force use of a keytab for computer account login */
+		adcli_conn_set_allowed_login_types (conn, ADCLI_LOGIN_COMPUTER_ACCOUNT);
+		ktname = adcli_enroll_get_keytab_name (enroll);
+		adcli_conn_set_login_keytab_name (conn, ktname ? ktname : "");
+	}
+
+	res = adcli_enroll_load (enroll);
+	if (res != ADCLI_SUCCESS) {
+		errx (-res, "couldn't lookup domain info from keytab: %s",
+		      adcli_get_last_error ());
+	}
+
+	res = adcli_conn_connect (conn);
+	if (res != ADCLI_SUCCESS) {
+		errx (-res, "couldn't connect to %s domain: %s",
+		      adcli_conn_get_domain_name (conn),
+		      adcli_get_last_error ());
+	}
+
+	res = adcli_enroll_update (enroll, flags);
+	if (res != ADCLI_SUCCESS) {
+		errx (-res, "updating membership with domain %s failed: %s",
+		      adcli_conn_get_domain_name (conn),
+		      adcli_get_last_error ());
+	}
+
+	if (details)
+		dump_details (conn, enroll, show_password);
+	else if (show_password)
+		dump_password (conn, enroll);
+
+	adcli_enroll_unref (enroll);
+
+	return 0;
+}
+
+
 int
 adcli_tool_computer_preset (adcli_conn *conn,
                             int argc,
@@ -429,11 +551,13 @@ adcli_tool_computer_preset (adcli_conn *conn,
 		case 'h':
 			adcli_tool_usage (options, usages);
 			adcli_tool_usage (options, common_usages);
+			adcli_enroll_unref (enroll);
 			return 0;
 		case '?':
 		case ':':
 			adcli_tool_usage (options, usages);
 			adcli_tool_usage (options, common_usages);
+			adcli_enroll_unref (enroll);
 			return 2;
 		default:
 			parse_option ((Option)opt, optarg, conn, enroll);
@@ -518,6 +642,7 @@ adcli_tool_computer_reset (adcli_conn *conn,
 		case ':':
 			adcli_tool_usage (options, usages);
 			adcli_tool_usage (options, common_usages);
+			adcli_enroll_unref (enroll);
 			return opt == 'h' ? 0 : 2;
 		default:
 			parse_option ((Option)opt, optarg, conn, enroll);
@@ -591,6 +716,7 @@ adcli_tool_computer_delete (adcli_conn *conn,
 		case ':':
 			adcli_tool_usage (options, usages);
 			adcli_tool_usage (options, common_usages);
+			adcli_enroll_unref (enroll);
 			return opt == 'h' ? 0 : 2;
 		default:
 			parse_option ((Option)opt, optarg, conn, enroll);
@@ -606,6 +732,12 @@ adcli_tool_computer_delete (adcli_conn *conn,
 
 	adcli_conn_set_allowed_login_types (conn, ADCLI_LOGIN_USER_ACCOUNT);
 
+	res = adcli_enroll_load (enroll);
+	if (res != ADCLI_SUCCESS) {
+		errx (-res, "couldn't lookup domain info from keytab: %s",
+		      adcli_get_last_error ());
+	}
+
 	res = adcli_conn_connect (conn);
 	if (res != ADCLI_SUCCESS) {
 		errx (-res, "couldn't connect to %s domain: %s",
